<?php
class WHMCS_License
{
    
    private $licensekey = "";
    private $localkey = "";
    private $keydata = array();
    private $salt = "";
    private $date = "";
    private $localkeydecoded = false;
    private $responsedata = "";
    private $postmd5hash = "";
    private $localkeydays = "10";
    private $allowcheckfaildays = "5";
    private $debuglog = array();
    private $version = "7a1bbff560de83ab800c4d1d2f215b91006be8e6";
    protected static $instance;
    /**
     * The WHMCS_License singleton
     *
     * @var WHMCS_License
     */
    public function __construct()
    {
        
        $whmcs            = WHMCS_Application::getinstance();
        $this->licensekey = $whmcs->get_license_key();
        $this->localkey   = $whmcs->get_config("License");
        $this->salt       = sha1('' . "WHMCS" . $whmcs->get_config("Version") . "TFB" . $whmcs->get_hash());
        $this->date       = date("Ymd");
        $this->decodeLocalOnce();
        if (isset($_GET["forceremote"])) {
            $this->forceRemoteCheck();
            WHMCS_Terminus::getinstance()->doExit();
        }
    }
    
    /**
     * Set the WHMCS_License singleton.
     *
     * @param WHMCS_License $license
     * @return WHMCS_License
     */
    protected static function setInstance($license)
    {
        
        self::$instance = $license;
        return $license;
    }
    
    /**
     * Remove the WHMCS_License singleton.
     */
    protected static function destroyInstance()
    {
        
        self::$instance = null;
        return;
    }
    
    /**
     * Retrieve a WHMCS_License object via singleton.
     *
     * @return WHMCS_License
     */
    public static function getInstance()
    {
        if (is_null(self::$instance)) {
            self::setinstance(new WHMCS_License());
        }
        return self::$instance;
    }
    
    /**
     * Retrieve a list of licensing server IPs
     *
     * @return array
     */
    private function getHosts()
    {
        
        $hosts = gethostbynamel("licensing28.whmcs.com");
        if ($hosts === false) {
            $hosts = array();
        }
        return $hosts;
    }
    
    private function getLicenseKey()
    {
        
        return $this->licensekey;
    }
    
    private function getHostIP()
    {
        
        if (isset($_SERVER["SERVER_ADDR"])) {
            $ip = $_SERVER["SERVER_ADDR"];
        } else {
            if (isset($_SERVER["LOCAL_ADDR"])) {
                $ip = $_SERVER["LOCAL_ADDR"];
            } else {
                if (function_exists("gethostname")) {
                    $ip = gethostbyname(gethostname());
                } else {
                    $ip = "";
                }
            }
        }
        return $ip;
    }
    
    private function getHostDomain()
    {
        
        return isset($_SERVER["SERVER_NAME"]) ? ($_SERVER["SERVER_NAME"]) : "";
    }
    
    private function getHostDir()
    {
        
        return ROOTDIR;
    }
    
    public function getSalt()
    {
        
        return $this->salt;
    }
    
    public function getDate()
    {
        
        return $this->date;
    }
    
    public function checkLocalKeyExpiry()
    {
        
        $originalcheckdate = $this->getKeyData("checkdate");
        $localexpirymax    = date("Ymd", mktime(0, 0, 0, date("m"), date("d") - $this->localkeydays, date("Y")));
        if ($originalcheckdate < $localexpirymax) {
            return false;
        }
        $localmax = date("Ymd", mktime(0, 0, 0, date("m"), date("d") + 2, date("Y")));
        if ($localmax < $originalcheckdate) {
            return false;
        }
        return true;
    }
    
    public function remoteCheck($forceRemote = false)
    {
        
        try {
            $localkeyvalid = $this->decodeLocalOnce();
            $this->debug('' . "Local Key Valid: " . $localkeyvalid);
            if ($localkeyvalid) {
                $localkeyvalid = $this->checkLocalKeyExpiry();
                $this->debug('' . "Local Key Expiry: " . $localkeyvalid);
                if ($localkeyvalid) {
                    $localkeyvalid = $this->validateLocalKey();
                    $this->debug('' . "Local Key Validation: " . $localkeyvalid);
                }
            }
            if (false || !$localkeyvalid || $forceRemote) {
                $whmcs                     = WHMCS_Application::getinstance();
                $results["status"]         = "Active";
                $results["key"]            = $this->licensekey;
                $results["registeredname"] = $whmcs->get_config("CompanyName");
                $results["productname"]    = "Owned License No Branding";
                $results["productid"]      = "5";
                $results["billingcycle"]   = "One Time";
                $results["validdomains"]   = $this->getHostDomain();
                $results["validips"]       = $this->getHostIP();
                $results["validdirs"]      = $this->getHostDir();
                $results["checkdate"]      = $this->getDate();
                $results["version"]        = $whmcs->getVersion()->getCanonical();
                $results["regdate"]        = "2014-02-06";
                $results["nextduedate"]    = "2050-02-06";
                $results["addons"]         = array(
                    array(
                        'name' => 'Branding Removal',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Support and Updates',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Project Management Addon',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Licensing Addon',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Mobile Edition',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'iPhone App',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Android App',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Configurable Package Addon',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    ),
                    array(
                        'name' => 'Live Chat Monthly No Branding',
                        'nextduedate' => '2050-12-30',
                        'status' => 'Active'
                    )
                );
                $this->setKeyData($results);
                $this->updateLocalKey();
            }
            $this->debug("Remote Check Done");
        }
        catch (WHMCS_Exception $exception) {
            $this->debug(sprintf("License Error: %s", $exception->getMessage()));
            return false;
        }
        return true;
    }
    
    private function getLocalMaxExpiryDate()
    {
        
        return date("Ymd", mktime(0, 0, 0, date("m"), date("d") - ($this->localkeydays + $this->allowcheckfaildays), date("Y")));
    }
    
    private function buildQuery($postfields)
    {
        
        $query_string = "";
        foreach ($postfields as $k => $v) {
            $query_string .= '' . $k . "=" . urlencode($v) . "&";
            continue;
        }
        return $query_string;
    }
    
    private function callHome($postfields)
    {
        
        $query_string = $this->buildQuery($postfields);
        $res          = $this->callHomeLoop($query_string, 5);
        if ($res) {
            return $res;
        }
        return $this->callHomeLoop($query_string, 30);
    }
    
    private function callHomeLoop($query_string, $timeout = 5)
    {
        
        $hostips = $this->getHosts();
        foreach ($hostips as $hostip) {
            $responsecode = $this->makeCall($hostip, $query_string, $timeout);
            if (!($responsecode == 200)) {
                continue;
            }
            return $this->responsedata;
        }
        return false;
    }
    
    private function makeCall($ip, $query_string, $timeout = 5)
    {
        
        $url = "https://" . $ip . "/license/verify53.php";
        $this->debug('' . "Request URL " . $url);
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_POST, 1);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $query_string);
        curl_setopt($ch, CURLOPT_TIMEOUT, $timeout);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        $this->responsedata = curl_exec($ch);
        $responsecode       = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $this->debug('' . "Response Code: " . $responsecode . " Data: " . $this->responsedata);
        if (curl_error($ch)) {
            $this->debug("Curl Error: " . curl_error($ch) . " - " . curl_errno($ch));
        }
        curl_close($ch);
        return $responsecode;
    }
    
    private function processResponse($data)
    {
        
        $data           = strrev($data);
        $data           = base64_decode($data);
        $results        = unserialize($data);
        $this->posthash = $results["hash"];
        unset($results["hash"]);
        $results["checkdate"] = $this->getDate();
        return $results;
    }
    
    private function updateLocalKey()
    {
        
        $data_encoded = serialize($this->keydata);
        $data_encoded = base64_encode($data_encoded);
        $data_encoded = sha1($this->getDate() . $this->getSalt()) . $data_encoded;
        $data_encoded = strrev($data_encoded);
        $splpt        = strlen($data_encoded) / 2;
        $data_encoded = substr($data_encoded, $splpt) . substr($data_encoded, 0, $splpt);
        $data_encoded = sha1($data_encoded . $this->getSalt()) . $data_encoded . sha1($data_encoded . $this->getSalt() . time());
        $data_encoded = base64_encode($data_encoded);
        $data_encoded = wordwrap($data_encoded, 80, "\n", true);
        WHMCS_Application::getinstance()->set_config("License", $data_encoded);
        $this->debug("Updated Local Key");
        return null;
    }
    
    public function forceRemoteCheck()
    {
        
        $this->remoteCheck(true);
        return null;
    }
    
    private function setInvalid($reason = "Invalid")
    {
        
        $this->keydata = array(
            "status" => $reason
        );
        return null;
    }
    
    private function decodeLocal()
    {
        
        $this->debug("Decoding local key");
        $localkey = $this->localkey;
        if (!$localkey) {
            return false;
        }
        $localkey  = str_replace("\n", "", $localkey);
        $localkey  = base64_decode($localkey);
        $localdata = substr($localkey, 40, 0 - 40);
        $md5hash   = substr($localkey, 0, 40);
        if ($md5hash == sha1($localdata . $this->getSalt())) {
            $splpt             = strlen($localdata) / 2;
            $localdata         = substr($localdata, $splpt) . substr($localdata, 0, $splpt);
            $localdata         = strrev($localdata);
            $md5hash           = substr($localdata, 0, 40);
            $localdata         = substr($localdata, 40);
            $localdata         = base64_decode($localdata);
            $localkeyresults   = unserialize($localdata);
            $originalcheckdate = $localkeyresults["checkdate"];
            if ($md5hash == sha1($originalcheckdate . $this->getSalt())) {
                if (isset($localkeyresults["key"])) {
                    if (isset($localkeyresults["key"]) || $localkeyresults["key"] == WHMCS_Application::getinstance()->get_license_key()) {
                        $this->debug("Local Key Decode Successful");
                        $this->setKeyData($localkeyresults);
                    }
                } else {
                    $this->debug("License Key Invalid");
                }
            } else {
                $this->debug("Local Key MD5 Hash 2 Invalid");
            }
        } else {
            $this->debug("Local Key MD5 Hash Invalid");
        }
        $this->localkeydecoded = true;
        return $this->getKeyData("status") == "Active" ? (true) : (false);
    }
    
    private function decodeLocalOnce()
    {
        
        if ($this->localkeydecoded) {
            return true;
        }
        return $this->decodeLocal();
    }
    
    private function isRunningInCLI()
    {
        
        if (php_sapi_name() == "cli");
        return php_sapi_name() == "cli" || empty($_SERVER["REMOTE_ADDR"]);
    }
    
    private function validateLocalKey()
    {
        
        if ($this->getKeyData("status") != "Active") {
            $this->debug("Local Key Status Check Failure");
            return false;
        }
        if ($this->isRunningInCLI()) {
            $this->debug("Running in CLI Mode");
        } else {
            $this->debug("Running in Browser Mode");
            if ($this->isValidDomain($this->getHostDomain())) {
                $this->debug("Domain Validated Successfully");
            } else {
                $this->debug("Local Key Domain Check Failure");
                return false;
            }
            $ip = $this->getHostIP();
            $this->debug('' . "Host IP Address: " . $ip);
            if (!$ip) {
                $this->debug("IP Could Not Be Determined - Skipping Local Validation of IP");
            } else {
                if (!trim($this->getKeyData("validips"))) {
                    $this->debug("No Valid IPs returned by license check - Cloud Based License - Skipping Local Validation of IP");
                } else {
                    if ($this->isValidIP($ip)) {
                        $this->debug("IP Validated Successfully");
                    } else {
                        $this->debug("Local Key IP Check Failure");
                        return false;
                    }
                }
            }
        }
        if ($this->isValidDir($this->getHostDir())) {
            $this->debug("Directory Validated Successfully");
        } else {
            $this->debug("Local Key Directory Check Failure");
            return false;
        }
        return true;
    }
    
    private function isValidDomain($domain)
    {
        
        $validdomains = $this->getArrayKeyData("validdomains");
        return in_array($domain, $validdomains);
    }
    
    private function isValidIP($ip)
    {
        
        $validips = $this->getArrayKeyData("validips");
        return in_array($ip, $validips);
    }
    
    private function isValidDir($dir)
    {
        
        $validdirs = $this->getArrayKeyData("validdirs");
        return in_array($dir, $validdirs);
    }
    
    private function revokeLocal()
    {
        
        WHMCS_Application::getinstance()->set_config("License", "");
        return null;
    }
    
    public function getKeyData($var)
    {
        
        return isset($this->keydata[$var]) ? ($this->keydata[$var]) : "";
    }
    
    private function setKeyData($data)
    {
        
        $this->keydata = $data;
        return null;
    }
    
    /**
     * Retrieve a license element as an array, that would otherwise be a
     * delimited string
     *
     * NOTE: use of this method should be very limited. New license elements
     * added to the license data should strongly consider not depending on the
     * use of this function, but instead structure the data and let the
     * transmission layer do the serialize/unserialize
     *
     * @param string $var License data element whose value is a comma delimited string
     *
     * @return array
     * @throws WHMCS_Exception when internal license key data structure is not
     * as expected
     */
    protected function getArrayKeyData($var)
    {
        
        $listData = array();
        $rawData  = $this->getKeyData($var);
        if (is_string($rawData)) {
            $listData = explode(",", $rawData);
            foreach ($listData as $k => $v) {
                if (is_string($v)) {
                    $listData[$k] = trim($v);
                    continue;
                }
                throw new WHMCS_Exception("Invalid license data structure");
                continue;
            }
        } else {
            if (!is_null($rawData)) {
                throw new WHMCS_Exception("Invalid license data structure");
            }
        }
        return $listData;
    }
    
    public function getRegisteredName()
    {
        
        return $this->getKeyData("registeredname");
    }
    
    public function getProductName()
    {
        
        return $this->getKeyData("productname");
    }
    
    public function getStatus()
    {
        
        return $this->getKeyData("status");
    }
    
    public function getSupportAccess()
    {
        
        return $this->getKeyData("supportaccess");
    }
    
    /**
     * Retrieve a list of Addons as known by the license
     * 
     * @return array
     */
    protected function getLicensedAddons()
    {
        
        $licensedAddons = $this->getKeyData("addons");
        if (!is_array($licensedAddons)) {
            $licensedAddons = array();
        }
        return $licensedAddons;
    }
    
    public function getActiveAddons()
    {
        
        $licensedAddons = $this->getLicensedAddons();
        $activeAddons   = array();
        foreach ($licensedAddons as $addon) {
            if (!($addon["status"] == "Active")) {
                continue;
            }
            $activeAddons[] = $addon["name"];
            continue;
        }
        return $activeAddons;
    }
    
    public function isActiveAddon($addon)
    {
        
        return in_array($addon, $this->getActiveAddons()) ? (true) : (false);
    }
    
    public function getExpiryDate($showday = false)
    {
        
        $expiry = $this->getKeyData("nextduedate");
        if (!$expiry) {
            $expiry = "Never";
        } else {
            if ($showday) {
                $expiry = date("l, jS F Y", strtotime($expiry));
            } else {
                $expiry = date("jS F Y", strtotime($expiry));
            }
        }
        return $expiry;
    }
    
    /**
     * Get a version object that will represent the latest publicly available version
     *
     * If the licensing API does not return a valid version number for
     * whatever reason, it assumes latest version = installed version
     * to allow application to continue un-affected
     *
     * @return WHMCS_Version_SemanticVersion
     */
    public function getLatestPublicVersion()
    {
        
        try {
            $latestVersion = new WHMCS_Version_SemanticVersion($this->getKeyData("latestpublicversion"));
        }
        catch (WHMCS_Exception_Version_BadVersionNumber $e) {
            $whmcs         = WHMCS_Application::getinstance();
            $latestVersion = $whmcs->getVersion();
        }
        return $latestVersion;
    }
    
    /**
     * Get a version object that will represent the latest available pre-release version
     *
     * If the licensing API does not return a valid version number for
     * whatever reason, it assumes latest version = installed version
     * to allow application to continue un-affected
     *
     * @return WHMCS_Version_SemanticVersion
     */
    public function getLatestPreReleaseVersion()
    {
        
        try {
            $latestVersion = new WHMCS_Version_SemanticVersion($this->getKeyData("latestprereleaseversion"));
        }
        catch (WHMCS_Exception_Version_BadVersionNumber $e) {
            $whmcs         = WHMCS_Application::getinstance();
            $latestVersion = $whmcs->getVersion();
        }
        return $latestVersion;
    }
    
    /**
     * Get a version object that will represent the latest appropriate version based on current installation
     *
     * If running a pre-release (beta/rc) it returns the latest pre-release version
     * Otherwise it returns the latest publicly available version
     *
     * @return WHMCS_Version_SemanticVersion
     */
    public function getLatestVersion()
    {
        
        $whmcs            = WHMCS_Application::getinstance();
        $installedVersion = $whmcs->getVersion();
        if (in_array($installedVersion->getPreReleaseIdentifier(), array(
            "beta",
            "rc"
        ))) {
            $latestVersion = $this->getLatestPreReleaseVersion();
        } else {
            $latestVersion = $this->getLatestPublicVersion();
        }
        return $latestVersion;
    }
    
    /**
     * Determines if an update is available for the currently installed files
     *
     * @throws WHMCS_Exception_Version_BadVersionNumber If version number invalid
     *
     * @return bool
     */
    public function isUpdateAvailable()
    {
        
        $whmcs            = WHMCS_Application::getinstance();
        $installedVersion = $whmcs->getVersion();
        $latestVersion    = $this->getLatestVersion();
        return WHMCS_Version_SemanticVersion::compare($latestVersion, $installedVersion, ">");
    }
    
    private function getRequiresUpdates()
    {
        
        return $this->getKeyData("requiresupdates") ? (true) : (false);
    }
    
    public function checkOwnedUpdates()
    {
        
        if (!$this->getRequiresUpdates()) {
            return true;
        }
        $whmcs          = WHMCS_Application::getinstance();
        $licensedAddons = $this->getLicensedAddons();
        foreach ($licensedAddons as $addon) {
            if ($addon["name"] == "Support and Updates") {
                if (!($addon["name"] == "Support and Updates" || $addon["status"] == "Active")) {
                    continue;
                }
                if (!(str_replace("-", "", $whmcs->getReleaseDate()) < str_replace("-", "", $addon["nextduedate"]))) {
                    continue;
                }
                return true;
            }
            continue;
        }
        return false;
    }
    
    public function getBrandingRemoval()
    {
        
        if (in_array($this->getProductName(), array(
            "Owned License No Branding",
            "Monthly Lease No Branding"
        ))) {
            return true;
        }
        $licensedAddons = $this->getLicensedAddons();
        foreach ($licensedAddons as $addon) {
            if ($addon["name"] == "Branding Removal") {
                if (!($addon["name"] == "Branding Removal" || $addon["status"] == "Active")) {
                    continue;
                }
                return true;
            }
            continue;
        }
        return false;
    }
    
    public function getVersionHash()
    {
        
        return $this->version;
    }
    
    private function debug($msg)
    {
        
        $this->debuglog[] = $msg;
        return null;
    }
    
    /**
     * Retrieve all errors
     *
     * @return array
     */
    public function getDebugLog()
    {
        
        return $this->debuglog;
    }
    
}